var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _AbstractGossipLog_instances, _AbstractGossipLog_validatePayload, _AbstractGossipLog_verifySignature, _AbstractGossipLog_apply, _AbstractGossipLog_insert;
import { TypedEventEmitter, CustomEvent } from "@libp2p/interface";
import { logger } from "@libp2p/logger";
import { bytesToHex as hex } from "@noble/hashes/utils";
import { Ed25519DelegateSigner } from "@canvas-js/signatures";
import { assert } from "@canvas-js/utils";
import { Mempool } from "./Mempool.js";
import { Driver } from "./sync/driver.js";
import { decodeId, encodeId, encodeSignedMessage, getNextClock, messageIdPattern, MIN_MESSAGE_ID, MAX_MESSAGE_ID, decodeSignedMessage, } from "./schema.js";
import { topicPattern } from "./utils.js";
export class AbstractGossipLog extends TypedEventEmitter {
    constructor(init) {
        super();
        _AbstractGossipLog_instances.add(this);
        this.mempool = new Mempool();
        _AbstractGossipLog_validatePayload.set(this, void 0);
        _AbstractGossipLog_verifySignature.set(this, void 0);
        _AbstractGossipLog_apply.set(this, void 0);
        assert(topicPattern.test(init.topic), "invalid topic (must match [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        this.indexAncestors = init.indexAncestors ?? false;
        this.signer = init.signer ?? new Ed25519DelegateSigner();
        __classPrivateFieldSet(this, _AbstractGossipLog_apply, init.apply, "f");
        __classPrivateFieldSet(this, _AbstractGossipLog_validatePayload, init.validatePayload ?? ((payload) => true), "f");
        __classPrivateFieldSet(this, _AbstractGossipLog_verifySignature, init.verifySignature ?? this.signer.verify, "f");
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async replay() {
        await this.read(async (txn) => {
            for await (const [key, value] of txn.messages.entries()) {
                const [id, signature, message] = this.decode(value);
                assert(id === decodeId(key), "expected id === decodeId(key)");
                await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(txn, [id, signature, message]);
            }
        });
    }
    async *iterate(lowerBound = null, upperBound = null, options = {}) {
        const { id: lowerId, inclusive: lowerInclusive } = lowerBound ?? { id: MIN_MESSAGE_ID, inclusive: true };
        const { id: upperId, inclusive: upperInclusive } = upperBound ?? { id: MAX_MESSAGE_ID, inclusive: true };
        assert(messageIdPattern.test(lowerId), "lowerBound.id: invalid message ID");
        assert(messageIdPattern.test(upperId), "upperBound.id: invalid message ID");
        for await (const [key, value] of this.entries({ key: encodeId(lowerId), inclusive: lowerInclusive }, { key: encodeId(upperId), inclusive: upperInclusive }, options)) {
            const [id, signature, message] = this.decode(value);
            assert(id === decodeId(key), "expected id === decodeId(key)");
            yield [id, signature, message];
        }
    }
    encode(signature, message) {
        assert(this.topic === message.topic, "expected this.topic === topic");
        assert(__classPrivateFieldGet(this, _AbstractGossipLog_validatePayload, "f").call(this, message.payload), "error encoding message (invalid payload)");
        return encodeSignedMessage(signature, message);
    }
    decode(value) {
        const [id, signature, { topic, clock, parents, payload }] = decodeSignedMessage(value);
        assert(this.topic === topic, "expected this.topic === topic");
        assert(__classPrivateFieldGet(this, _AbstractGossipLog_validatePayload, "f").call(this, payload), "error decoding message (invalid payload)");
        return [id, signature, { topic, clock, parents, payload }];
    }
    async getClock() {
        const heads = await this.read((txn) => txn.getHeads());
        const clock = getNextClock(heads);
        return [clock, heads.map(decodeId)];
    }
    async has(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        return await this.read(({ messages }) => messages.get(encodeId(id)) !== null);
    }
    async get(id) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const value = await this.read(({ messages }) => messages.get(encodeId(id)));
        if (value === null) {
            return [null, null];
        }
        const [_, signature, message] = this.decode(value);
        return [signature, message];
    }
    /**
     * Sign and append a new *unsigned* message to the end of the log.
     * The currently unmerged heads of the local log are used as parents.
     */
    async append(payload, options = {}) {
        const signer = options.signer ?? this.signer;
        const { id, signature, message, result, root } = await this.write(async (txn) => {
            const heads = await txn.getHeads();
            const clock = getNextClock(heads);
            const parents = heads.map(decodeId);
            const message = { topic: this.topic, clock, parents, payload };
            const signature = await signer.sign(message);
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("appending message %s: %O", id, message);
            const result = await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, id, signature, message, [key, value]);
            const root = await txn.messages.getRoot();
            return { id, signature, message, result, root };
        });
        this.dispatchEvent(new CustomEvent("commit", { detail: { topic: this.topic, root } }));
        this.log("commited root %s", hex(root.hash));
        return { id, signature, message, result };
    }
    /**
     * Insert an existing signed message into the log (ie received via PubSub).
     * If any of the parents are not present, insert the message into the mempool instead.
     */
    async insert(signature, message) {
        await __classPrivateFieldGet(this, _AbstractGossipLog_verifySignature, "f").call(this, signature, message);
        const { id, root } = await this.write(async (txn) => {
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("inserting message %s", id);
            const missingParents = new Set();
            this.log("looking up %s parents", message.parents.length);
            for (const parentId of message.parents) {
                // TODO: txn.messages.getMany
                const parent = await txn.messages.get(encodeId(parentId));
                if (parent === null) {
                    this.log("missing parent %s", parentId);
                    missingParents.add(parentId);
                }
                else {
                    this.log("found parent %s", parentId);
                }
            }
            if (missingParents.size > 0) {
                this.log("missing %d/%d parents", missingParents.size, message.parents.length);
                this.mempool.add(id, { signature, message }, missingParents);
                return { id };
            }
            await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, id, signature, message, [key, value]);
            const root = await txn.messages.getRoot();
            return { id, root };
        });
        if (root !== undefined) {
            this.dispatchEvent(new CustomEvent("commit", { detail: { topic: this.topic, root } }));
            this.log("commited root %s", hex(root.hash));
        }
        return { id };
    }
    async getAncestors(id, atOrBefore) {
        assert(messageIdPattern.test(id), "invalid message ID");
        const results = new Set();
        await this.read((txn) => txn.getAncestors(encodeId(id), atOrBefore, results));
        this.log("getAncestors of %s atOrBefore %d: %o", id, atOrBefore, results);
        return Array.from(results).sort();
    }
    async isAncestor(id, ancestor) {
        assert(messageIdPattern.test(id), "invalid message ID");
        assert(messageIdPattern.test(ancestor), "invalid message ID");
        return await this.read((txn) => txn.isAncestor(encodeId(id), encodeId(ancestor)));
    }
    /**
     * Sync with a remote source, applying and inserting all missing messages into the local log
     */
    async sync(source, options = {}) {
        let messageCount = 0;
        const start = performance.now();
        const root = await this.write(async (txn) => {
            const driver = new Driver(this.topic, source, txn.messages);
            for await (const [key, value] of driver.sync()) {
                const [id, signature, message] = this.decode(value);
                assert(id === decodeId(key), "expected id === decodeId(key)");
                await __classPrivateFieldGet(this, _AbstractGossipLog_verifySignature, "f").call(this, signature, message);
                const existingMessage = await txn.messages.get(key);
                if (existingMessage === null) {
                    for (const parent of message.parents) {
                        // TODO: txn.messages.getMany
                        const parentKey = encodeId(parent);
                        const parentValue = await txn.messages.get(parentKey);
                        if (parentValue === null) {
                            this.log.error("missing parent %s of message %s: %O", parent, id, message);
                            if (this.indexAncestors) {
                                throw new Error(`missing parent ${parent} of message ${id}`);
                            }
                            else {
                                continue; // don't try to insert, just skip and try to get the message on the next sync
                            }
                        }
                    }
                    await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, id, signature, message, [key, value]);
                    if (options.timeoutController)
                        options.timeoutController.delay();
                    messageCount++;
                }
            }
            return await txn.messages.getRoot();
        }, options);
        const duration = Math.ceil(performance.now() - start);
        const peer = options.sourceId;
        this.dispatchEvent(new CustomEvent("sync", { detail: { peer, duration, messageCount } }));
        this.dispatchEvent(new CustomEvent("commit", { detail: { root } }));
        this.log("commited root %s", hex(root.hash));
        return { root, messageCount };
    }
    /**
     * Serve a read-only snapshot of the merkle tree
     */
    async serve(callback, options = {}) {
        await this.read((txn) => callback(txn.messages), options);
    }
}
_AbstractGossipLog_validatePayload = new WeakMap(), _AbstractGossipLog_verifySignature = new WeakMap(), _AbstractGossipLog_apply = new WeakMap(), _AbstractGossipLog_instances = new WeakSet(), _AbstractGossipLog_insert = async function _AbstractGossipLog_insert(txn, id, signature, message, [key, value] = this.encode(signature, message)) {
    this.log("applying %s %O", id, message);
    let result;
    try {
        result = await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(txn, [id, signature, message]);
    }
    catch (error) {
        this.dispatchEvent(new CustomEvent("error", { detail: { error } }));
        throw error;
    }
    this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message, result } }));
    await txn.insert(id, signature, message, [key, value]);
    for (const [childId, { signature, message }] of this.mempool.observe(id)) {
        await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, childId, signature, message);
    }
    return result;
};
