import PQueue from "p-queue";
import pDefer from "p-defer";
import { equals } from "uint8arrays";
import { MemoryTree, MemoryStore } from "@canvas-js/okra-memory";
import { assert } from "@canvas-js/utils";
import { KEY_LENGTH, encodeId, encodeSignedMessage } from "../schema.js";
import { AbstractGossipLog } from "../AbstractGossipLog.js";
import { SyncDeadlockError, cborNull } from "../utils.js";
import { getAncestors, indexAncestors, isAncestor } from "../ancestors.js";
export class GossipLog extends AbstractGossipLog {
    static async open(init) {
        const messages = await MemoryTree.open();
        return new GossipLog(messages, init);
    }
    constructor(messages, init) {
        super(init);
        this.messages = messages;
        this.heads = new MemoryStore();
        this.ancestors = new MemoryStore();
        this.queue = new PQueue({ concurrency: 1 });
        this.incomingSyncPeers = new Set();
        this.outgoingSyncPeers = new Set();
    }
    async close() {
        this.log("closing");
        this.queue.clear();
        await this.queue.onIdle();
        await this.messages.store.close();
        await this.heads.close();
    }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        const deferred = pDefer();
        this.queue.add(() => {
            return deferred.promise;
        });
        try {
            for await (const node of this.messages.nodes(0, lowerBound ?? { key: null, inclusive: false }, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            deferred.resolve();
        }
    }
    async read(callback, options = {}) {
        const targetId = options.targetId ?? null;
        if (targetId !== null) {
            if (this.outgoingSyncPeers.has(targetId)) {
                throw new SyncDeadlockError(`deadlock with peer ${targetId}`);
            }
        }
        const result = await this.queue.add(async () => {
            if (targetId !== null) {
                this.incomingSyncPeers.add(targetId);
            }
            try {
                return await callback({
                    getHeads: () => getHeads(this.heads),
                    getAncestors: async (key, atOrBefore, results) => getAncestors(this.ancestors, key, atOrBefore, results),
                    isAncestor: (key, ancestorKey, visited = new Set()) => isAncestor(this.ancestors, key, ancestorKey, visited),
                    messages: this.messages,
                });
            }
            catch (err) {
                this.log.error("error in transaction: %O", err);
            }
            finally {
                if (targetId !== null) {
                    this.incomingSyncPeers.delete(targetId);
                }
            }
        });
        return result;
    }
    async write(callback, options = {}) {
        const sourceId = options.sourceId ?? null;
        if (sourceId !== null) {
            if (this.incomingSyncPeers.has(sourceId)) {
                throw new SyncDeadlockError(`deadlock with peer ${sourceId}`);
            }
        }
        const result = await this.queue.add(async () => {
            if (sourceId !== null) {
                this.outgoingSyncPeers.add(sourceId);
            }
            try {
                return await callback({
                    getHeads: () => getHeads(this.heads),
                    getAncestors: async (key, atOrBefore, results) => getAncestors(this.ancestors, key, atOrBefore, results),
                    isAncestor: (key, ancestorKey, visited = new Set()) => isAncestor(this.ancestors, key, ancestorKey, visited),
                    insert: async (id, signature, message, [key, value] = encodeSignedMessage(signature, message)) => {
                        await this.messages.set(key, value);
                        const parentKeys = message.parents.map(encodeId);
                        await this.heads.set(key, cborNull);
                        for (const parentKey of parentKeys) {
                            await this.heads.delete(parentKey);
                        }
                        if (this.indexAncestors) {
                            await indexAncestors(this.ancestors, key, parentKeys);
                        }
                    },
                    messages: this.messages,
                });
            }
            catch (err) {
                this.log.error("error in transaction: %O", err);
                throw err;
            }
            finally {
                if (sourceId !== null) {
                    this.outgoingSyncPeers.delete(sourceId);
                }
            }
        });
        return result;
    }
}
async function getHeads(heads) {
    const parents = [];
    for await (const [key, value] of heads.entries()) {
        assert(key.byteLength === KEY_LENGTH, "internal error (expected key.byteLength === KEY_LENGTH)");
        assert(equals(value, cborNull), "internal error (unexpected parent entry value)");
        parents.push(key);
    }
    return parents;
}
