import fs from "node:fs";
import { equals } from "uint8arrays";
import { Environment, Transaction, Tree } from "@canvas-js/okra-node";
import { assert } from "@canvas-js/utils";
import { KEY_LENGTH, encodeId, encodeSignedMessage } from "../schema.js";
import { AbstractGossipLog } from "../AbstractGossipLog.js";
import { getAncestors, indexAncestors, isAncestor } from "../ancestors.js";
import { cborNull } from "../utils.js";
export class GossipLog extends AbstractGossipLog {
    static async open(init, path) {
        if (!fs.existsSync(path)) {
            fs.mkdirSync(path, { recursive: true });
        }
        const env = new Environment(path, { databases: 3 });
        const gossipLog = new GossipLog(env, init);
        await gossipLog.write(async (txn) => { });
        return gossipLog;
    }
    constructor(env, init) {
        super(init);
        this.env = env;
    }
    async close() {
        this.log("closing");
        await this.env.close();
    }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        const txn = new Transaction(this.env, { readOnly: true });
        try {
            const tree = new Tree(txn, "messages");
            for await (const node of tree.nodes(0, lowerBound ?? { key: null, inclusive: false }, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            txn.abort();
        }
    }
    async read(callback) {
        this.log("opening read-only transaction");
        return await this.env.read(async (txn) => {
            const messages = new Tree(txn, "messages");
            const heads = txn.database("heads");
            const ancestors = txn.database("ancestors");
            return await callback({
                getHeads: () => getHeads(heads),
                getAncestors: async (key, atOrBefore, results) => getAncestors(ancestors, key, atOrBefore, results),
                isAncestor: (key, ancestorKey, visited = new Set()) => isAncestor(ancestors, key, ancestorKey, visited),
                messages,
            });
        });
    }
    async write(callback) {
        this.log("opening read-write transaction");
        return await this.env.write(async (txn) => {
            const messages = new Tree(txn, "messages");
            const heads = txn.database("heads");
            const ancestors = txn.database("ancestors");
            return await callback({
                getHeads: () => getHeads(heads),
                getAncestors: async (key, atOrBefore, results) => getAncestors(ancestors, key, atOrBefore, results),
                isAncestor: (key, ancestorKey, visited = new Set()) => isAncestor(ancestors, key, ancestorKey, visited),
                insert: async (id, signature, message, [key, value] = encodeSignedMessage(signature, message)) => {
                    messages.set(key, value);
                    const parentKeys = message.parents.map(encodeId);
                    heads.set(key, cborNull);
                    for (const parentKey of parentKeys) {
                        heads.delete(parentKey);
                    }
                    if (this.indexAncestors) {
                        await indexAncestors(ancestors, key, parentKeys);
                    }
                },
                messages,
            });
        });
    }
}
async function getHeads(heads) {
    const parents = [];
    for await (const [key, value] of heads.entries()) {
        assert(key.byteLength === KEY_LENGTH, "internal error (expected key.byteLength === KEY_LENGTH)");
        assert(equals(value, cborNull), "internal error (unexpected parent entry value)");
        parents.push(key);
    }
    return parents;
}
