import PQueue from "p-queue";
import pDefer from "p-defer";
import * as cbor from "@ipld/dag-cbor";
import pg from "pg";
import { hexToBytes } from "@noble/hashes/utils";
import { equals } from "uint8arrays";
import { PostgresTree, PostgresStore } from "@canvas-js/okra-pg";
import { assert } from "@canvas-js/utils";
import { KEY_LENGTH, decodeId, encodeId, encodeSignedMessage } from "../schema.js";
import { AbstractGossipLog } from "../AbstractGossipLog.js";
import { cborNull, getAncestorClocks } from "../utils.js";
import { getAncestorsSql } from "./get_ancestors.sql.js";
import { isAncestorSql } from "./is_ancestor.sql.js";
import { decodeClockSql } from "./decode_clock.sql.js";
import { pgCborSql } from "./pg_cbor.sql.js";
import { insertSql } from "./insert_updating_ancestors.sql.js";
import { insertMessageRemovingHeadsSql } from "./insert_message_removing_heads.sql.js";
import { decodeClock } from "../clock.js";
const initSql = [
    getAncestorsSql,
    isAncestorSql,
    decodeClockSql,
    pgCborSql,
    insertSql,
    insertMessageRemovingHeadsSql,
].join("\n");
async function getAncestors(log, key, atOrBefore) {
    const result = await log.ancestorsClient.query(`SELECT ret_results FROM get_ancestors($1, $2, '{}'::bytea[]);`, [key, atOrBefore]);
    const { rows } = result;
    const row = rows[0];
    return row.ret_results;
}
async function isAncestor(log, key, ancestorKey) {
    const { rows } = await log.ancestorsClient.query(`SELECT ret_result FROM is_ancestor($1, $2, '{}'::bytea[]);`, [
        key,
        ancestorKey,
    ]);
    const row = rows[0];
    return row.ret_result;
}
async function insertUpdatingAncestors(client, ancestors, key, parentKeys) {
    const [clock] = decodeClock(key);
    const ancestorClocks = Array.from(getAncestorClocks(clock));
    const { rows } = await client.query(`SELECT insert_updating_ancestors($1, $2::bytea[], $3::integer[]);`, [key, parentKeys.map(Buffer.from), ancestorClocks]);
    assert(rows.length > 0);
    const [{ insert_updating_ancestors: result }] = rows;
    const ancestorLinks = result.map((arr) => arr.map((id) => hexToBytes(id.replace("\\x", ""))));
    await ancestors.set(key, cbor.encode(ancestorLinks));
}
async function insertMessageRemovingHeads(log, key, value, hash, cborNull, heads) {
    const args = [key, value, hash, cborNull, heads.map(Buffer.from)];
    await log.ancestorsClient.query(`CALL insert_message_removing_heads($1, $2, $3, $4, $5::bytea[]);`, args);
}
export class GossipLog extends AbstractGossipLog {
    static async open(init, connectionConfig, clear = false) {
        const pool = typeof connectionConfig === "string"
            ? new pg.Pool({ connectionString: connectionConfig })
            : new pg.Pool(connectionConfig);
        const messagesClient = await pool.connect();
        const headsClient = await pool.connect();
        const ancestorsClient = await pool.connect();
        pool.on("error", (err, client) => {
            console.error("Unexpected error on idle client", err);
        });
        const messages = await PostgresTree.initialize(messagesClient, { prefix: this.MESSAGES_TABLE_PREFIX, clear });
        const heads = await PostgresStore.initialize(headsClient, { table: this.HEADS_TABLE, clear });
        const ancestors = await PostgresStore.initialize(ancestorsClient, { table: this.ANCESTORS_TABLE, clear });
        await ancestorsClient.query(initSql);
        const gossipLog = new GossipLog({
            pool,
            messagesClient,
            headsClient,
            ancestorsClient,
            messages,
            heads,
            ancestors,
        }, init);
        await gossipLog.write(async (txn) => { });
        return gossipLog;
    }
    constructor({ pool, messagesClient, headsClient, ancestorsClient, messages, heads, ancestors, }, init) {
        super(init);
        this.queue = new PQueue({ concurrency: 1 });
        this.pool = pool;
        this.messagesClient = messagesClient;
        this.headsClient = headsClient;
        this.ancestorsClient = ancestorsClient;
        this.messages = messages;
        this.heads = heads;
        this.ancestors = ancestors;
    }
    async close() {
        this.log("closing");
        this.queue.clear();
        await this.queue.onIdle();
        this.messagesClient.release();
        this.headsClient.release();
        this.ancestorsClient.release();
        await this.pool.end();
    }
    async *entries(lowerBound = null, upperBound = null, options = {}) {
        const deferred = pDefer();
        this.queue.add(() => {
            return deferred.promise;
        });
        try {
            for await (const node of this.messages.nodes(0, lowerBound ?? null, upperBound, options)) {
                assert(node.key !== null, "expected node.key !== null");
                assert(node.value !== undefined, "expected node.value !== undefined");
                yield [node.key, node.value];
            }
        }
        finally {
            deferred.resolve();
        }
    }
    async read(callback) {
        this.log("opening read-only transaction");
        const result = await this.queue.add(async () => {
            // console.log("start read tx")
            const result = await callback({
                getHeads: async () => getHeads(this.headsClient),
                getAncestors: (key, atOrBefore, results) => getAncestors(this, key, atOrBefore).then((keys) => keys.forEach((key) => results.add(decodeId(key)))),
                isAncestor: (key, ancestorKey) => isAncestor(this, key, ancestorKey),
                messages: this.messages,
            });
            // console.log("end read tx")
            return result;
        });
        return result;
    }
    async write(callback) {
        this.log("opening read-write transaction");
        const result = await this.queue.add(async () => {
            // console.log("start write tx")
            const result = await callback({
                getHeads: async () => getHeads(this.headsClient),
                getAncestors: (key, atOrBefore, results) => getAncestors(this, key, atOrBefore).then((keys) => keys.forEach((key) => results.add(decodeId(key)))),
                isAncestor: (key, ancestorKey) => isAncestor(this, key, ancestorKey),
                insert: async (id, signature, message, [key, value] = encodeSignedMessage(signature, message)) => {
                    const parentKeys = message.parents.map(encodeId);
                    const hash = this.messages.hashEntry(key, value);
                    await insertMessageRemovingHeads(this, key, value, hash, cborNull, parentKeys);
                    if (this.indexAncestors) {
                        await insertUpdatingAncestors(this.ancestorsClient, this.ancestors, key, parentKeys);
                    }
                },
                messages: this.messages,
            });
            // console.log("end write tx")
            return result;
        });
        return result;
    }
}
GossipLog.MESSAGES_TABLE_PREFIX = "messages";
GossipLog.HEADS_TABLE = "heads";
GossipLog.ANCESTORS_TABLE = "ancestors";
async function getHeads(heads) {
    const { rows } = await heads.query(`SELECT * FROM heads ORDER BY key`);
    return rows.map(({ key, value }) => {
        assert(key.byteLength === KEY_LENGTH, "internal error (expected key.byteLength === KEY_LENGTH)");
        assert(equals(value, cborNull), "internal error (unexpected parent entry value)");
        return key;
    });
}
